from __future__ import annotations
from .merkletree import MerkleTree
from .bloomfilter import BloomFilter
from .ecc import ECCCurve, ECPoint, KeyPair, ECCException, ecdsa_verify, ecdsa_sign
import hashlib

__all__ = ['MerkleTree', 'BloomFilter', 'ECCCurve', 'ECPoint', 'KeyPair', 'sign', 'verify_signature',
           'ECCException']


def sign(message: bytes, private_key: bytes, curve=ECCCurve.SECP256R1, hash_func=hashlib.sha256) -> bytes:
    return ecdsa_sign(private_key, message, curve, hash_func)


def verify_signature(message: bytes,
                     signature: bytes,
                     public_key: bytes,
                     curve: ECCCurve = ECCCurve.SECP256R1,
                     hash_func=hashlib.sha256) -> bool:
    """
    Test is the `signature` is signed by `public_key` valid for `message`.

    Args:
        message: the data the signature was created over
        signature: the signature to validate for `message`
        public_key: the public key the message was signed with
        curve: the ECC curve to use for verifying

    Raises:
        ValueError: for the Secp256r1 curve if the public key has an invalid format

    """
    pub_key = ECPoint(public_key, curve, True)
    return ecdsa_verify(signature, message, pub_key, hash_func)



import abc
import unittest
from neo3.core import types
from neo3 import storage, contracts
from neo3.network import payloads
from neo3.contracts import manifest
from contextlib import suppress
from copy import deepcopy


class AbstractBlockStorageTest(abc.ABC, unittest.TestCase):
    """
    A helper class to easily test backend specific code
    """

    @abc.abstractmethod
    def db_factory(self):
        """ Implement to return an instance of your DB """

    def test_rawview_bestblockheight(self):

        raw_view = self.db.get_rawview()
        self.assertEqual(-1, raw_view.best_block_height)

        raw_view.blocks.put(self.block1)
        self.assertEqual(1, raw_view.best_block_height)

        with self.assertRaises(AttributeError):
            raw_view.best_block_height = 2
        self.assertEqual(1, raw_view.best_block_height)

    def setUp(self) -> None:
        self.db = self.db_factory()
        signer = payloads.Signer(account=types.UInt160.from_string("d7678dd97c000be3f33e9362e673101bac4ca654"),
                                 scope=payloads.WitnessScope.NONE)
        tx = payloads.Transaction(version=0,
                                  nonce=123,
                                  system_fee=456,
                                  network_fee=789,
                                  valid_until_block=1,
                                  attributes=[],
                                  signers=[signer],
                                  script=b'\x01',
                                  witnesses=[payloads.Witness(invocation_script=b'', verification_script=b'\x55')])

        self.header = payloads.Header(version=0,
                                      prev_hash=types.UInt256.from_string("f782c7fbb2eef6afe629b96c0d53fb525eda64ce5345057caf975ac3c2b9ae0a"),
                                      timestamp=123,
                                      nonce=0,
                                      index=1,
                                      primary_index=0,
                                      next_consensus=types.UInt160.from_string("d7678dd97c000be3f33e9362e673101bac4ca654"),
                                      witness=payloads.Witness(invocation_script=b'', verification_script=b'\x55'),)
        self.block1 = payloads.Block(self.header, transactions=[tx])
        self.block1.rebuild_merkle_root()
        self.block1_hash = self.block1.hash()

        self.block2 = deepcopy(self.block1)
        self.block2.header.index = 2
        self.block2_hash = self.block2.hash()

    def test_raw(self):
        raw_view = self.db.get_rawview()

        # we should not find anything in an empty db
        target_block_hash = types.UInt256.zero()
        with self.assertRaises(KeyError):
            raw_view.blocks.get(target_block_hash)
        self.assertIsNone(raw_view.blocks.try_get(target_block_hash))

        with self.assertRaises(KeyError):
            raw_view.blocks.get_by_height(1)
        self.assertIsNone(raw_view.blocks.try_get_by_height(1))

        # fill the db
        raw_view.blocks.put(self.block1)
        # and test it is immediately added
        block_from_db = raw_view.blocks.try_get(self.block1_hash)
        self.assertIsNotNone(block_from_db)
        self.assertEqual(self.block1, block_from_db)

        # test again but get via height
        block_from_db = raw_view.blocks.try_get_by_height(self.block1.index)
        self.assertIsNotNone(block_from_db)
        self.assertEqual(self.block1, block_from_db)

        # test getting all blocks
        raw_view.blocks.put(self.block2)
        blocks = list(raw_view.blocks.all())
        self.assertEqual(2, len(blocks))
        self.assertIn(self.block1, blocks)
        self.assertIn(self.block2, blocks)

        # finally try removing a block
        raw_view.blocks.delete(self.block1_hash)
        self.assertIsNone(raw_view.blocks.try_get(self.block1_hash))

    def test_snapshot_basic_add_delete_get(self):
        # test basic add, delete, get and separation
        raw_view = self.db.get_rawview()
        snapshot_view = self.db.get_snapshotview()

        # we should not find anything in an empty db
        target_block_hash = types.UInt256.zero()
        with self.assertRaises(KeyError):
            snapshot_view.blocks.get(target_block_hash)
        self.assertIsNone(snapshot_view.blocks.try_get(target_block_hash))

        # same as previous but by height
        with self.assertRaises(KeyError):
            snapshot_view.blocks.get_by_height(1)
        self.assertIsNone(snapshot_view.blocks.try_get_by_height(1))

        # add item
        snapshot_view.blocks.put(self.block1)
        # real backend should not be affected until a commit is called
        self.assertIsNone(raw_view.blocks.try_get(self.block1_hash))
        self.assertIsNone(raw_view.blocks.try_get_by_height(self.block1.index))

        # persist to backend
        snapshot_view.commit()
        block_from_db = raw_view.blocks.try_get(self.block1_hash)
        # and validate
        self.assertIsNotNone(block_from_db)
        self.assertEqual(self.block1, block_from_db)

        # same but by height to validate the height -> hash -> block mapping
        block_from_db = raw_view.blocks.try_get_by_height(self.block1.index)
        self.assertIsNotNone(block_from_db)
        self.assertEqual(self.block1, block_from_db)

        # finally, try deleting
        # get a clean view with no cache
        snapshot_view = self.db.get_snapshotview()
        snapshot_view.blocks.delete(self.block1_hash)

        # real backend should still have it, snapshot not
        self.assertIsNotNone(raw_view.blocks.try_get(self.block1_hash))
        self.assertIsNone(snapshot_view.blocks.try_get(self.block1_hash))

        # persist and validate real backend also doesn't have it anymore
        snapshot_view.commit()
        self.assertIsNone(raw_view.blocks.try_get(self.block1_hash))

    def test_snapshot_add_duplicates(self):
        snapshot_view = self.db.get_snapshotview()

        # test double adding while already in cache
        snapshot_view.blocks.put(self.block1)
        with self.assertRaises(ValueError):
            snapshot_view.blocks.put(self.block1)

        # test double adding when not in cache, but in real backend
        snapshot_view.commit()
        # get a clean one with an empty cache
        snapshot_view = self.db.get_snapshotview()
        with self.assertRaises(ValueError):
            snapshot_view.blocks.put(self.block1)

    def test_snapshot_add_while_cache_marked_deleted(self):
        # an item can exist in the real backend, and be marked in cache to be deleted
        # it should then be possible to delete it from cache without exceptions

        # fill real backend
        raw_view = self.db.get_rawview()
        raw_view.blocks.put(self.block1)

        # ensure item is marked as deleted in cache
        snapshot_view = self.db.get_snapshotview()
        snapshot_view.blocks.delete(self.block1_hash)

        # now test by adding add the item again
        success = False
        with suppress(ValueError):
            snapshot_view.blocks.put(self.block1)
            success = True
        self.assertTrue(success)

    def test_snapshot_get_various(self):
        snapshot_view = self.db.get_snapshotview()

        # get non existing item
        with self.assertRaises(KeyError):
            snapshot_view.blocks.get(self.block1_hash)

        with self.assertRaises(KeyError):
            snapshot_view.blocks.get_by_height(self.block1.index)

        # test read only
        raw_view = self.db.get_rawview()
        raw_view.blocks.put(self.block1)
        block = snapshot_view.blocks.get(self.block1_hash, read_only=True)
        block.header.index = 123

        block_again = snapshot_view.blocks.get(self.block1_hash, read_only=True)
        # We validate the hash of the original with the hash of the block we retrieved.
        # The modification of the index attribute above changes the hash, if it persisted
        # the following test fails
        self.assertEqual(self.block1_hash, block_again.hash())

        # same as above but test read_only for get_by_height()
        block = snapshot_view.blocks.get_by_height(self.block1.index, read_only=True)
        block.header.index = 123
        block_again = snapshot_view.blocks.get(self.block1_hash, read_only=True)
        self.assertEqual(self.block1_hash, block_again.hash())

    def test_snapshot_clone_put(self):
        raw_view = self.db.get_rawview()
        snapshot_view = self.db.get_snapshotview()
        snapshot_view.blocks.put(self.block1)

        clone_view = snapshot_view.clone()
        # validate it has the same block
        self.assertEqual(self.block1, clone_view.blocks.try_get(self.block1_hash))

        # put some in the clone
        clone_view.blocks.put(self.block2)

        # validate it is not visible anywhere but in the clone
        self.assertIsNone(raw_view.blocks.try_get(self.block2_hash))
        self.assertIsNone(snapshot_view.blocks.try_get(self.block2_hash))
        self.assertEqual(self.block2, clone_view.blocks.try_get(self.block2_hash))

        # commit changes of the clone back into the snapshot
        # should affect only the snapshot, not the real backend
        clone_view.commit()
        self.assertIsNone(raw_view.blocks.try_get(self.block2_hash))
        self.assertEqual(self.block2, snapshot_view.blocks.try_get(self.block2_hash))

        # finally commit to real db
        snapshot_view.commit()
        self.assertEqual(self.block2, raw_view.blocks.try_get(self.block2_hash))

    def test_snapshot_clone_delete(self):
        raw_view = self.db.get_rawview()
        raw_view.blocks.put(self.block1)
        snapshot_view = self.db.get_snapshotview()

        # perform a get to fill the cache
        snapshot_view.blocks.get(self.block1_hash)

        clone_view = snapshot_view.clone()
        # now test deleting an item
        clone_view.blocks.delete(self.block1_hash)
        # test it's gone in the clone, but nowhere else
        self.assertIsNone(clone_view.blocks.try_get(self.block1_hash))
        self.assertIsNone(clone_view.blocks.try_get_by_height(self.block1.index))
        self.assertIsNotNone(snapshot_view.blocks.try_get(self.block1_hash))
        self.assertIsNotNone(snapshot_view.blocks.try_get_by_height(self.block1.index))
        self.assertIsNotNone(raw_view.blocks.try_get(self.block1_hash))
        self.assertIsNotNone(raw_view.blocks.try_get_by_height(self.block1.index))

        # commit the clone into the snapshot
        clone_view.commit()
        # and validate it is also gone in the snapshot but not the real db
        self.assertIsNone(snapshot_view.blocks.try_get(self.block1_hash))
        self.assertIsNotNone(raw_view.blocks.try_get(self.block1_hash))

        # finally persist to real db
        snapshot_view.commit()
        self.assertIsNone(raw_view.blocks.try_get(self.block1_hash))

    def test_snapshot_clone_update(self):
        # we currently have no way of testing for changes, as any attribute changes on the Block object
        # will change the hash and
        pass

    def test_all(self):
        raw_view = self.db.get_rawview()
        raw_view.blocks.put(self.block1)
        raw_view.blocks.put(self.block2)

        snapshot_view = self.db.get_snapshotview()

        # get() a block to fill the cache so we can test sorting and readonly behaviour
        # block1's hash comes before block2 when sorting. So we cache that first as the all() function internals
        # collect the results from the backend (=block2) before results from the cache (=block1).
        # Therefore if block1 is found in the first position of the all() results, we can
        # conclude that the sort() happened correctly.
        snapshot_view.blocks.get(self.block1_hash)
        blocks = list(snapshot_view.blocks.all())
        self.assertEqual(2, len(blocks))
        self.assertEqual(self.block1, blocks[0])
        self.assertEqual(self.block2, blocks[1])

        # ensure all() results are readonly
        blocks[0].transactions.append(payloads.Transaction._serializable_init())
        blocks[1].transactions.append(payloads.Transaction._serializable_init())

        block1_from_snap = snapshot_view.blocks.get(self.block1_hash, read_only=True)
        block2_from_snap = snapshot_view.blocks.get(self.block2_hash, read_only=True)
        self.assertNotEqual(3, len(block1_from_snap.transactions))
        self.assertNotEqual(3, len(block2_from_snap.transactions))

        # test clone all()
        block3 = deepcopy(self.block1)
        block3.header.index = 3

        clone_view = snapshot_view.clone()
        clone_view.blocks.put(block3)
        blocks = list(clone_view.blocks.all())
        self.assertEqual(3, len(blocks))
        self.assertEqual(2, len(list(snapshot_view.blocks.all())))
        self.assertEqual(self.block1, blocks[0])
        self.assertEqual(self.block2, blocks[2])
        self.assertEqual(block3, blocks[1])

    def test_snapshot_bestblockheight(self):
        snapshot_view = self.db.get_snapshotview()
        self.assertEqual(-1, snapshot_view.best_block_height)

        snapshot_view.best_block_height = 2
        self.assertEqual(2, snapshot_view.best_block_height)

        # nothing yet in raw view
        raw_view = self.db.get_rawview()
        self.assertEqual(-1, raw_view.best_block_height)

        clone_view = snapshot_view.clone()
        self.assertEqual(2, clone_view.best_block_height)

        clone_view.best_block_height = 3
        self.assertEqual(3, clone_view.best_block_height)
        self.assertEqual(2, snapshot_view.best_block_height)
        self.assertEqual(-1, raw_view.best_block_height)

        clone_view.commit()
        self.assertEqual(3, snapshot_view.best_block_height)
        self.assertEqual(-1, raw_view.best_block_height)

        snapshot_view.commit()
        self.assertEqual(3, raw_view.best_block_height)

    def test_snapshot_bestblockheight_2(self):
        snapshot_view = self.db.get_snapshotview()
        snapshot_view.blocks.put(self.block1)
        snapshot_view.commit()

        raw_view = self.db.get_rawview()
        self.assertEqual(self.block1.index, raw_view.best_block_height)


class AbstractContractStorageTest(abc.ABC, unittest.TestCase):
    """
    A helper class to easily test backend specific code
    """

    @abc.abstractmethod
    def db_factory(self):
        """ Implement to return an instance of your DB """

    def shortDescription(self):
        # disable docstring printing in test runner
        return None

    def setUp(self) -> None:
        self.db = self.db_factory()
        dummy_method = contracts.ContractMethodDescriptor("test1", 0, [], contracts.ContractParameterType.VOID, True)
        nef1 = contracts.NEF(script=b'\x01\x02')
        self.contract1 = contracts.ContractState(1, nef1, contracts.ContractManifest(), 0, types.UInt160(b'\x01' * 20))
        # the ABI must have at least 1 method or sanity checks will fail
        self.contract1.manifest.abi.methods.append(dummy_method)
        self.contract1_hash = self.contract1.hash
        nef2 = contracts.NEF(script=b'\x03\x04')
        self.contract2 = contracts.ContractState(2, nef2, contracts.ContractManifest(), 0, types.UInt160(b'\x02' * 20))
        self.contract2.manifest.abi.methods.append(dummy_method)
        self.contract2_hash = self.contract2.hash
        nef3 = contracts.NEF(script=b'\x03\x04')
        self.contract3 = contracts.ContractState(3, nef3, contracts.ContractManifest(), 0, types.UInt160(b'\x03' * 20))
        self.contract3.manifest.abi.methods.append(dummy_method)
        self.contract3_hash = self.contract3.hash

    def test_raw(self):
        raw_view = self.db.get_rawview()

        # we should not find anything in an empty db
        target_contract_hash = types.UInt160.zero()
        with self.assertRaises(KeyError):
            raw_view.contracts.get(target_contract_hash)
        self.assertIsNone(raw_view.contracts.try_get(target_contract_hash))

        # fill the db
        raw_view.contracts.put(self.contract1)
        # and test it is immediately added
        contract_from_db = raw_view.contracts.try_get(self.contract1_hash)
        self.assertIsNotNone(contract_from_db)
        self.assertEqual(self.contract1, contract_from_db)

        # test getting all contracts
        raw_view.contracts.put(self.contract2)
        contracts = list(raw_view.contracts.all())
        self.assertEqual(2, len(contracts))
        self.assertIn(self.contract1, contracts)
        self.assertIn(self.contract2, contracts)

        # finally try removing the contract
        raw_view.contracts.delete(self.contract1_hash)
        self.assertIsNone(raw_view.contracts.try_get(self.contract1_hash))

    def test_snapshot_basic_add_delete_get(self):
        # test basic add, delete, get and separation
        raw_view = self.db.get_rawview()
        snapshot_view = self.db.get_snapshotview()

        # we should not find anything in an empty db
        target_contract_hash = types.UInt160.zero()
        with self.assertRaises(KeyError) as context:
            snapshot_view.contracts.get(target_contract_hash)
        self.assertIsNone(snapshot_view.contracts.try_get(target_contract_hash))

        # add item
        snapshot_view.contracts.put(self.contract1)
        # real backend should not be affected until a commit is called
        self.assertIsNone(raw_view.contracts.try_get(self.contract1_hash))

        # persist to backend
        snapshot_view.commit()
        contract_from_db = raw_view.contracts.try_get(self.contract1_hash)
        # and validate
        self.assertIsNotNone(contract_from_db)
        self.assertEqual(self.contract1, contract_from_db)

        # finally, try deleting
        # get a clean view with no cache
        snapshot_view = self.db.get_snapshotview()
        snapshot_view.contracts.delete(self.contract1_hash)

        # real backend should still have it, snapshot not
        self.assertIsNotNone(raw_view.contracts.try_get(self.contract1_hash))
        self.assertIsNone(snapshot_view.contracts.try_get(self.contract1_hash))

        # persist and validate real backend also doesn't have it anymore
        snapshot_view.commit()
        self.assertIsNone(raw_view.contracts.try_get(self.contract1_hash))

    def test_snapshot_add_duplicates(self):
        snapshot_view = self.db.get_snapshotview()

        # test double adding while already in cache
        snapshot_view.contracts.put(self.contract1)
        with self.assertRaises(ValueError):
            snapshot_view.contracts.put(self.contract1)

        # test double adding when not in cache, but in real backend
        snapshot_view.commit()
        # get a clean one with an empty cache
        snapshot_view = self.db.get_snapshotview()
        with self.assertRaises(ValueError):
            snapshot_view.contracts.put(self.contract1)

    def test_snapshot_add_while_cache_marked_deleted(self):
        # an item can exist in the real backend, and be marked in cache to be deleted
        # it should then be possible to delete it from cache without exceptions

        # fill real backend
        raw_view = self.db.get_rawview()
        raw_view.contracts.put(self.contract1)

        # ensure item is marked as deleted in cache
        snapshot_view = self.db.get_snapshotview()
        snapshot_view.contracts.delete(self.contract1_hash)

        # now test by adding add the item again
        success = False
        with suppress(ValueError):
            snapshot_view.contracts.put(self.contract1)
            success = True
        self.assertTrue(success)

    def test_snapshot_delete_various(self):
        snapshot_view = self.db.get_snapshotview()

        # delete non existing item with empty cache should throw no errors
        ok = False
        with suppress(Exception):
            snapshot_view.contracts.delete(self.contract1_hash)
            ok = True
        self.assertTrue(ok)

        # delete an item that was only added to the cache
        snapshot_view.contracts.put(self.contract1)
        snapshot_view.contracts.delete(self.contract1_hash)
        # we test if the cache is empty by looking at the count of items in the cache
        self.assertEqual(0, len(snapshot_view.contracts._dictionary))

        # finally we get an existing item, which puts it into CHANGED state
        # then DELETE it and verify it is marked as deleted

        # start by filling the real backend
        raw_view = self.db.get_rawview()
        raw_view.contracts.put(self.contract1)

        # get the item so it creates a cache entry
        contract_from_db = snapshot_view.contracts.get(self.contract1_hash)
        self.assertIsNotNone(contract_from_db)
        # now delete it
        snapshot_view.contracts.delete(self.contract1_hash)
        # and validate the item in cache is marked as DELETED
        self.assertEqual(1, len(snapshot_view.contracts._dictionary))
        trackable = snapshot_view.contracts._dictionary.get(self.contract1_hash, None)  # type: storage.Trackable
        self.assertIsNotNone(trackable)
        self.assertEqual(storage.TrackState.DELETED, trackable.state)

    def test_snapshot_get_various(self):
        snapshot_view = self.db.get_snapshotview()

        # get non existing item
        with self.assertRaises(KeyError):
            snapshot_view.contracts.get(self.contract1_hash)

        # test read only
        raw_view = self.db.get_rawview()
        raw_view.contracts.put(self.contract1)
        contract = snapshot_view.contracts.get(self.contract1_hash, read_only=True)
        contract.script = b'\x11\x22'

        contract_again = snapshot_view.contracts.get(self.contract1_hash, read_only=True)
        # We validate the hash of the original with the hash of the contract we retrieved.
        # The modification of the script attribute above changes the hash, if it persisted
        # the following test fails
        self.assertEqual(self.contract1_hash, contract_again.hash)

    def test_snapshot_clone_put(self):
        raw_view = self.db.get_rawview()
        snapshot_view = self.db.get_snapshotview()
        snapshot_view.contracts.put(self.contract1)

        clone_view = snapshot_view.clone()
        # validate it has the same tx
        self.assertEqual(self.contract1, clone_view.contracts.try_get(self.contract1_hash))

        # put some in the clone
        clone_view.contracts.put(self.contract2)

        # validate it is not visible anywhere but in the clone
        self.assertIsNone(raw_view.contracts.try_get(self.contract2_hash))
        self.assertIsNone(snapshot_view.contracts.try_get(self.contract2_hash))
        self.assertEqual(self.contract2, clone_view.contracts.try_get(self.contract2_hash))

        # commit changes of the clone back into the snapshot
        # should affect only the snapshot, not the real backend
        clone_view.commit()
        self.assertIsNone(raw_view.contracts.try_get(self.contract2_hash))
        self.assertEqual(self.contract2, snapshot_view.contracts.try_get(self.contract2_hash))

        # finally commit to real db
        snapshot_view.commit()
        self.assertEqual(self.contract2, raw_view.contracts.try_get(self.contract2_hash))

    def test_snapshot_clone_delete(self):
        raw_view = self.db.get_rawview()
        raw_view.contracts.put(self.contract1)
        snapshot_view = self.db.get_snapshotview()

        # perform a get to fill the cache
        snapshot_view.contracts.get(self.contract1_hash)

        clone_view = snapshot_view.clone()
        # now test deleting an item
        clone_view.contracts.delete(self.contract1_hash)
        # test it's gone in the clone, but nowhere else
        self.assertIsNone(clone_view.contracts.try_get(self.contract1_hash))
        self.assertIsNotNone(snapshot_view.contracts.try_get(self.contract1_hash))
        self.assertIsNotNone(raw_view.contracts.try_get(self.contract1_hash))

        # commit the clone into the snapshot
        clone_view.commit()
        # and validate it is also gone in the snapshot but not the real db
        self.assertIsNone(snapshot_view.contracts.try_get(self.contract1_hash))
        self.assertIsNotNone(raw_view.contracts.try_get(self.contract1_hash))

        # finally persist to real db
        snapshot_view.commit()
        self.assertIsNone(raw_view.contracts.try_get(self.contract1_hash))

    def test_snapshot_clone_update(self):
        raw_view = self.db.get_rawview()
        raw_view.contracts.put(self.contract1)
        snapshot_view = self.db.get_snapshotview()

        # perform a get to fill the cache
        snapshot_view.contracts.get(self.contract1_hash)

        clone_view = snapshot_view.clone()
        contract_from_clone = clone_view.contracts.get(self.contract1_hash)  # type: contracts.ContractState
        # modify one of the attributes
        contract_from_clone.manifest.extra = True

        # validate the snapshot and real backend are not affected
        contract_from_snapshot = snapshot_view.contracts.get(self.contract1_hash)
        contract_from_real_db = raw_view.contracts.get(self.contract1_hash)
        self.assertNotEqual(True, contract_from_snapshot.manifest.extra)
        self.assertNotEqual(True, contract_from_real_db.manifest.extra)

        # commit clone
        clone_view.commit()
        # now snapshot should be updated, but real db not
        contract_from_snapshot = snapshot_view.contracts.get(self.contract1_hash)
        contract_from_real_db = raw_view.contracts.get(self.contract1_hash)
        self.assertEqual(True, contract_from_snapshot.manifest.extra)
        self.assertNotEqual(True, contract_from_real_db.manifest.extra)

        # finally persist to real db
        snapshot_view.commit()
        contract_from_real_db = raw_view.contracts.get(self.contract1_hash)
        self.assertEqual(True, contract_from_real_db.manifest.extra)

    def test_all(self):
        raw_view = self.db.get_rawview()
        raw_view.contracts.put(self.contract1)
        raw_view.contracts.put(self.contract2)

        snapshot_view = self.db.get_snapshotview()
        # get contract 2 to add a cache entry, so we can confirm correct sorting
        snapshot_view.contracts.get(self.contract2_hash, read_only=True)

        contracts = list(snapshot_view.contracts.all())
        self.assertEqual(2, len(contracts))
        self.assertEqual(self.contract1, contracts[0])
        self.assertEqual(self.contract2, contracts[1])

        # Ensure results are readonly. We modify the manifest, because that's the attribute which does not affect the
        # key the contract is stored under (a.k.a does not affect the contract script hash)
        mani = manifest.ContractManifest()
        mani._attr_for_test = 111
        contracts[0].manifest = mani
        contracts[1].manifest = mani

        mani1_from_snapshot = snapshot_view.contracts.get(self.contract1_hash, read_only=True)
        mani2_from_snapshot = snapshot_view.contracts.get(self.contract2_hash, read_only=True)
        # validate the manifest is unchanged
        self.assertIsNone(mani1_from_snapshot.manifest.extra)
        self.assertIsNone(mani2_from_snapshot.manifest.extra)

        # find something that's only in a clone
        clone_view = snapshot_view.clone()
        clone_view.contracts.put(self.contract3)

        contracts = list(clone_view.contracts.all())
        self.assertEqual(3, len(contracts))
        self.assertEqual(2, len(list(snapshot_view.contracts.all())))
        self.assertEqual(self.contract1, contracts[0])
        self.assertEqual(self.contract2, contracts[1])
        self.assertEqual(self.contract3, contracts[2])


class AbstractStorageStorageTest(abc.ABC, unittest.TestCase):
    """
    A helper class to easily test backend specific code
    """

    @abc.abstractmethod
    def db_factory(self):
        """ Implement to return an instance of your DB """

    def setUp(self) -> None:
        self.db = self.db_factory()
        self.storage1_id = 1

        self.storagekey1 = storage.StorageKey(self.storage1_id, b'\x01\x02')
        self.storageitem1 = storage.StorageItem(b'\x01\x01')

        self.storage2_id = 2

        self.storagekey2 = storage.StorageKey(self.storage2_id, b'\x03\x04')
        self.storageitem2 = storage.StorageItem(b'\x02\x02')

        self.storagekey3 = storage.StorageKey(self.storage2_id, b'\x03\x05')
        self.storageitem3 = storage.StorageItem(b'\x03\x03')

        self.storagekey4 = storage.StorageKey(self.storage2_id, b'\x04\x04')
        self.storageitem4 = storage.StorageItem(b'\x04\x04')

    def test_raw(self):
        raw_view = self.db.get_rawview()

        # we should not find any key in an empty db
        with self.assertRaises(KeyError):
            raw_view.storages.get(self.storagekey1)
        self.assertIsNone(raw_view.storages.try_get(self.storagekey1))

        # fill the db
        raw_view.storages.put(self.storagekey1, self.storageitem1)
        # and test it is immediately added
        item_from_db = raw_view.storages.try_get(self.storagekey1)
        self.assertIsNotNone(item_from_db)
        self.assertEqual(self.storageitem1, item_from_db)

        # test getting all storages
        raw_view.storages.put(self.storagekey2, self.storageitem2)
        storage_pairs = dict(raw_view.storages.all())
        self.assertEqual(2, len(storage_pairs))
        self.assertIn(self.storagekey1, storage_pairs)
        self.assertIn(self.storagekey2, storage_pairs)
        self.assertIn(self.storageitem1, storage_pairs.values())
        self.assertIn(self.storageitem2, storage_pairs.values())

        # test finding keys
        raw_view.storages.put(self.storagekey3, self.storageitem3)

        find_prefix = storage.create_find_prefix(self.storage2_id, b'\x03')
        storage_pairs = dict(raw_view.storages.find(find_prefix))
        self.assertEqual(2, len(storage_pairs))
        self.assertNotIn(self.storagekey1, storage_pairs)
        self.assertIn(self.storagekey2, storage_pairs)
        self.assertIn(self.storagekey3, storage_pairs)
        self.assertNotIn(self.storageitem1, storage_pairs.values())
        self.assertIn(self.storageitem2, storage_pairs.values())
        self.assertIn(self.storageitem3, storage_pairs.values())

        # finally try removing the data
        raw_view.storages.delete(self.storagekey1)
        self.assertIsNone(raw_view.storages.try_get(self.storagekey1))

    def test_snapshot_basic_add_delete_get(self):
        # test basic add, delete, get and separation
        raw_view = self.db.get_rawview()
        snapshot_view = self.db.get_snapshotview()

        # we should not find any key in an empty db
        with self.assertRaises(KeyError):
            raw_view.storages.get(self.storagekey1)
        self.assertIsNone(raw_view.storages.try_get(self.storagekey1))

        # add item
        snapshot_view.storages.put(self.storagekey1, self.storageitem1)
        # real backend should not be affected until a commit is called
        self.assertIsNone(raw_view.storages.try_get(self.storagekey1))

        # persist to backend
        snapshot_view.commit()
        storage_from_db = raw_view.storages.try_get(self.storagekey1)
        # and validate
        self.assertIsNotNone(storage_from_db)
        self.assertEqual(self.storageitem1, storage_from_db)

        # finally, try deleting
        # get a clean view with no cache
        snapshot_view = self.db.get_snapshotview()
        snapshot_view.storages.delete(self.storagekey1)

        # real backend should still have it, snapshot not
        self.assertIsNotNone(raw_view.storages.try_get(self.storagekey1))
        self.assertIsNone(snapshot_view.storages.try_get(self.storagekey1))

        # persist and validate real backend also doesn't have it anymore
        snapshot_view.commit()
        self.assertIsNone(raw_view.storages.try_get(self.storagekey1))

    def test_snapshot_with_context_manager(self):
        # partially repeat test_snapshot_basic_add_delete_get()
        raw_view = self.db.get_rawview()
        with self.db.get_snapshotview() as snapshot_view:
            # we should not find any key in an empty db
            with self.assertRaises(KeyError):
                raw_view.storages.get(self.storagekey1)
            self.assertIsNone(raw_view.storages.try_get(self.storagekey1))

            # add item
            snapshot_view.storages.put(self.storagekey1, self.storageitem1)
            # real backend should not be affected until a commit is called
            self.assertIsNone(raw_view.storages.try_get(self.storagekey1))

            # persist to backend
            snapshot_view.commit()
            storage_from_db = raw_view.storages.try_get(self.storagekey1)
            # and validate
            self.assertIsNotNone(storage_from_db)
            self.assertEqual(self.storageitem1, storage_from_db)

    def test_snapshot_add_duplicates(self):
        snapshot_view = self.db.get_snapshotview()

        # test double adding while already in cache
        snapshot_view.storages.put(self.storagekey1, self.storageitem1)
        with self.assertRaises(ValueError):
            snapshot_view.storages.put(self.storagekey1, self.storageitem1)

        # test double adding when not in cache, but in real backend
        snapshot_view.commit()
        # get a clean one with an empty cache
        snapshot_view = self.db.get_snapshotview()
        with self.assertRaises(ValueError):
            snapshot_view.storages.put(self.storagekey1, self.storageitem1)

    def test_snapshot_add_while_cache_marked_deleted(self):
        # an item can exist in the real backend, and be marked in cache to be deleted
        # it should then be possible to delete it from cache without exceptions

        # fill real backend
        raw_view = self.db.get_rawview()
        raw_view.storages.put(self.storagekey1, self.storageitem1)

        # ensure item is marked as deleted in cache
        snapshot_view = self.db.get_snapshotview()
        snapshot_view.storages.delete(self.storagekey1)

        # now test by adding add the item again
        success = False
        with suppress(ValueError):
            snapshot_view.storages.put(self.storagekey1, self.storageitem1)
            success = True
        self.assertTrue(success)

    def test_snapshot_delete_various(self):
        snapshot_view = self.db.get_snapshotview()

        # delete non existing item with empty cache should throw no errors
        ok = False
        with suppress(Exception):
            snapshot_view.storages.delete(self.storagekey1)
            ok = True
        self.assertTrue(ok)

        # delete an item that was only added to the cache
        snapshot_view.storages.put(self.storagekey1, self.storageitem1)
        snapshot_view.storages.delete(self.storagekey1)
        # we test if the cache is empty by looking at the count of items in the cache
        self.assertEqual(0, len(snapshot_view.storages._dictionary))

        # finally we get an existing item, which puts it into CHANGED state
        # then DELETE it and verify it is marked as deleted

        # start by filling the real backend
        raw_view = self.db.get_rawview()
        raw_view.storages.put(self.storagekey1, self.storageitem1)

        # get the item so it creates a cache entry
        value_from_db = snapshot_view.storages.get(self.storagekey1)
        self.assertIsNotNone(value_from_db)
        # now delete it
        snapshot_view.storages.delete(self.storagekey1)
        # and validate the item in cache is marked as DELETED
        self.assertEqual(1, len(snapshot_view.storages._dictionary))
        trackable = snapshot_view.storages._dictionary.get(self.storagekey1, None)  # type: storage.Trackable
        self.assertIsNotNone(trackable)
        self.assertEqual(storage.TrackState.DELETED, trackable.state)

    def test_snapshot_get_various(self):
        snapshot_view = self.db.get_snapshotview()

        # get non existing item
        with self.assertRaises(KeyError):
            snapshot_view.storages.get(self.storagekey1)

        # test read only
        raw_view = self.db.get_rawview()
        raw_view.storages.put(self.storagekey1, self.storageitem1)
        storageitem = snapshot_view.storages.get(self.storagekey1, read_only=True)
        storageitem.value = b'\x55\x55'

        storageitem_again = snapshot_view.storages.get(self.storagekey1, read_only=True)
        self.assertEqual(self.storageitem1, storageitem_again)

    def test_snapshot_clone_put(self):
        raw_view = self.db.get_rawview()
        snapshot_view = self.db.get_snapshotview()
        snapshot_view.storages.put(self.storagekey1, self.storageitem1)

        clone_view = snapshot_view.clone()
        # validate it has the same storage item
        self.assertEqual(self.storageitem1, clone_view.storages.try_get(self.storagekey1))

        # put some in the clone
        clone_view.storages.put(self.storagekey2, self.storageitem2)

        # validate it is not visible anywhere but in the clone
        self.assertIsNone(raw_view.storages.try_get(self.storagekey2))
        self.assertIsNone(snapshot_view.storages.try_get(self.storagekey2))
        self.assertEqual(self.storageitem2, clone_view.storages.try_get(self.storagekey2))

        # commit changes of the clone back into the snapshot
        # should affect only the snapshot, not the real backend
        clone_view.commit()
        self.assertIsNone(raw_view.storages.try_get(self.storagekey2))
        self.assertEqual(self.storageitem2, snapshot_view.storages.try_get(self.storagekey2))

        # finally commit to real db
        snapshot_view.commit()
        self.assertEqual(self.storageitem2, raw_view.storages.try_get(self.storagekey2))

    def test_snapshot_clone_delete(self):
        raw_view = self.db.get_rawview()
        raw_view.storages.put(self.storagekey1, self.storageitem1)
        snapshot_view = self.db.get_snapshotview()

        # perform a get to fill the cache
        snapshot_view.storages.get(self.storagekey1)

        clone_view = snapshot_view.clone()
        # now test deleting an item
        clone_view.storages.delete(self.storagekey1)
        # test it's gone in the clone, but nowhere else
        self.assertIsNone(clone_view.storages.try_get(self.storagekey1))
        self.assertIsNotNone(snapshot_view.storages.try_get(self.storagekey1))
        self.assertIsNotNone(raw_view.storages.try_get(self.storagekey1))

        # commit the clone into the snapshot
        clone_view.commit()
        # and validate it is also gone in the snapshot but not the real db
        self.assertIsNone(snapshot_view.storages.try_get(self.storagekey1))
        self.assertIsNotNone(raw_view.storages.try_get(self.storagekey1))

        # finally persist to real db
        snapshot_view.commit()
        self.assertIsNone(raw_view.storages.try_get(self.storagekey1))

    def test_snapshot_clone_update(self):
        raw_view = self.db.get_rawview()
        raw_view.storages.put(self.storagekey1, self.storageitem1)
        snapshot_view = self.db.get_snapshotview()

        # perform a get to fill the cache
        snapshot_view.storages.get(self.storagekey1)

        clone_view = snapshot_view.clone()
        value_from_clone = clone_view.storages.get(self.storagekey1)  # type: storage.StorageItem
        value_from_clone.value = b'\x55\x55'

        # validate the snapshot and real backend are not affected
        value_from_snapshot = snapshot_view.storages.get(self.storagekey1)
        value_from_real_db = raw_view.storages.get(self.storagekey1)
        self.assertNotEqual(b'\x55\x55', value_from_snapshot.value)
        self.assertNotEqual(b'\x55\x55', value_from_real_db.value)

        # commit clone
        clone_view.commit()
        # now snapshot should be updated, but real db not
        value_from_snapshot = snapshot_view.storages.get(self.storagekey1)
        value_from_real_db = raw_view.storages.get(self.storagekey1)
        self.assertEqual(b'\x55\x55', value_from_snapshot.value)
        self.assertNotEqual(b'\x55\x55', value_from_real_db.value)

        # finally persist to real db
        snapshot_view.commit()
        value_from_real_db = raw_view.storages.get(self.storagekey1)
        self.assertEqual(b'\x55\x55', value_from_real_db.value)

    def test_all(self):
        raw_view = self.db.get_rawview()
        raw_view.storages.put(self.storagekey1, self.storageitem1)
        raw_view.storages.put(self.storagekey2, self.storageitem2)
        raw_view.storages.put(self.storagekey3, self.storageitem3)

        snapshot_view = self.db.get_snapshotview()
        # get a key to fill the cache so we can test sorting and readonly behaviour
        # key3 should come after key 2
        snapshot_view.storages.get(self.storagekey3, read_only=True)

        all_pairs = dict(snapshot_view.storages.all(self.storage2_id))
        self.assertEqual(2, len(all_pairs))
        self.assertNotIn(self.storagekey1, all_pairs)
        self.assertEqual(self.storagekey2, list(all_pairs.keys())[0])
        self.assertEqual(self.storagekey3, list(all_pairs.keys())[1])
        self.assertNotIn(self.storageitem1, all_pairs.values())
        self.assertEqual(self.storageitem2, list(all_pairs.values())[0])
        self.assertEqual(self.storageitem3, list(all_pairs.values())[1])

        # test results are readonly by modifying the results and requesting it again from the snapshot.
        storage_item2 = list(all_pairs.values())[0]
        storage_item3 = list(all_pairs.values())[1]
        storage_item2.value = b'\x55\x55'
        storage_item3.value = b'\x55\x55'
        item2_from_snap = snapshot_view.storages.get(self.storagekey2, read_only=True)
        item3_from_snap = snapshot_view.storages.get(self.storagekey3, read_only=True)
        self.assertNotEqual(b'\x55\x55', item2_from_snap.value)
        self.assertNotEqual(b'\x55\x55', item3_from_snap.value)

        clone_view = snapshot_view.clone()
        clone_view.storages.put(self.storagekey4, self.storageitem4)

        all_pairs = dict(clone_view.storages.all(self.storage2_id))
        self.assertEqual(3, len(all_pairs))
        self.assertEqual(2, len(list(snapshot_view.storages.all(self.storage2_id))))
        self.assertNotIn(self.storagekey1, all_pairs)
        self.assertEqual(self.storagekey2, list(all_pairs.keys())[0])
        self.assertEqual(self.storagekey3, list(all_pairs.keys())[1])
        self.assertEqual(self.storagekey4, list(all_pairs.keys())[2])
        self.assertNotIn(self.storageitem1, all_pairs.values())
        self.assertEqual(self.storageitem2, list(all_pairs.values())[0])
        self.assertEqual(self.storageitem3, list(all_pairs.values())[1])
        self.assertEqual(self.storageitem4, list(all_pairs.values())[2])

    def test_find(self):
        raw_view = self.db.get_rawview()
        raw_view.storages.put(self.storagekey1, self.storageitem1)
        raw_view.storages.put(self.storagekey2, self.storageitem2)
        raw_view.storages.put(self.storagekey3, self.storageitem3)
        raw_view.storages.put(self.storagekey4, self.storageitem4)

        snapshot_view = self.db.get_snapshotview()

        # get a key to fill the cache so we can test sorting and readonly behaviour
        # key3 should come after key 2
        snapshot_view.storages.get(self.storagekey3, read_only=True)

        # key2,3 and 4 are of the same smart contract
        # only key2 and key3 start with \x03
        find_prefix_storage2 = storage.create_find_prefix(self.storage2_id, b'\x03')
        all_pairs = list(snapshot_view.storages.find(find_prefix_storage2))
        keys = list(map(lambda i: i[0], all_pairs))
        items = list(map(lambda i: i[1], all_pairs))
        self.assertEqual(2, len(all_pairs))
        self.assertNotIn(self.storagekey1, keys)
        self.assertEqual(self.storagekey2, keys[0])
        self.assertEqual(self.storagekey3, keys[1])
        self.assertNotIn(self.storagekey4, keys)

        self.assertNotIn(self.storageitem1, items)
        self.assertEqual(self.storageitem2, items[0])
        self.assertEqual(self.storageitem3, items[1])
        self.assertNotIn(self.storageitem4, items)

        # test for read only results
        items[0].value = b'\x55\x55'
        item_from_snap = snapshot_view.storages.get(self.storagekey3, read_only=True)
        self.assertNotEqual(b'\x55\x55', item_from_snap.value)

        keys[0].key = b'\x77\x77'

        # create a storage key that should match the above modification
        modified_key = storage.StorageKey(self.storage2_id, b'\x77\x77')
        # and we should not find it
        self.assertIsNone(snapshot_view.storages.try_get(modified_key, read_only=True))

        # test find in clone
        clone_key = storage.StorageKey(self.storage2_id, b'\x03\x88')
        clone_item = storage.StorageItem(b'\x99\x99')
        clone_view = snapshot_view.clone()
        clone_view.storages.put(clone_key, clone_item)
        all_pairs = list(clone_view.storages.find(find_prefix_storage2))
        self.assertEqual(3, len(all_pairs))

    def test_find_extra(self):
        """
        This test used to fail on Neo 2.x and was reported here: https://github.com/neo-project/neo/issues/946

        They changed the logic to be deterministic, so we simulate the same

        public class Test : DataCache<StorageKey, StorageItem>
        {
            static StorageItem si = new StorageItem { IsConstant = false, Value = new byte[] { 0x1 } };
            static StorageKey key1 = new StorageKey { ScriptHash = UInt160.Zero, Key = new byte[] { 1 } };
            static StorageKey key2 = new StorageKey { ScriptHash = UInt160.Zero, Key = new byte[] { 2 } };
            static StorageKey key3 = new StorageKey { ScriptHash = UInt160.Zero, Key = new byte[] { 3 } };

            static Dictionary<StorageKey, StorageItem> innerDictionary = new Dictionary<StorageKey, StorageItem>() {
                { key1, si },
                { key2, si },
                { key3, si }
            };

            public override void DeleteInternal(StorageKey key)
            {
                throw new NotImplementedException();
            }

            protected override void AddInternal(StorageKey key, StorageItem value)
            {
                throw new NotImplementedException();
            }

            protected override IEnumerable<KeyValuePair<StorageKey, StorageItem>> FindInternal(byte[] key_prefix)
            {
                foreach (var pair in innerDictionary)
                {
                    yield return new KeyValuePair<StorageKey, StorageItem>(pair.Key, pair.Value);
                }
            }

            protected override StorageItem GetInternal(StorageKey key)
            {
                return innerDictionary[key];
            }

            protected override StorageItem TryGetInternal(StorageKey key)
            {
                return innerDictionary[key];
            }

            protected override void UpdateInternal(StorageKey key, StorageItem value)
            {
                throw new NotImplementedException();
            }
        }

        public static void Main(string[] args)
        {
            DataCache<StorageKey, StorageItem> test = new Test();
            StorageKey key1 = new StorageKey { ScriptHash = UInt160.Zero, Key = new byte[] { 1 } };

            test.TryGet(key1);
            foreach (KeyValuePair<StorageKey, StorageItem> pair in test.Find())
            {
                Console.WriteLine($"{BitConverter.ToString(pair.Key.Key)}");
            }
        }


        """
        key1 = storage.StorageKey(0, b'key1')
        key2 = storage.StorageKey(0, b'key2')
        key3 = storage.StorageKey(0, b'key3')
        value1 = storage.StorageItem(b'value1')

        snapshot_view = self.db.get_snapshotview()
        snapshot_view.storages.put(key1, value1)
        snapshot_view.storages.put(key2, value1)
        snapshot_view.storages.put(key3, value1)
        results = list(snapshot_view.storages.all())
        self.assertEqual(key1, results[0][0])
        self.assertEqual(key2, results[1][0])
        self.assertEqual(key3, results[2][0])

        # NEO-cli sorts keys based on the serialized TKey value, we make 1 special case where we change the contract
        key1 = storage.StorageKey(1, key=b'key1')
        snapshot_view = self.db.get_snapshotview()
        snapshot_view.storages.put(key1, value1)
        snapshot_view.storages.put(key2, value1)
        snapshot_view.storages.put(key3, value1)
        results = list(snapshot_view.storages.all())
        self.assertEqual(key2, results[0][0])
        self.assertEqual(key3, results[1][0])
        self.assertEqual(key1, results[2][0])

    def test_issue_1672(self):
        # test if we are affected by https://github.com/neo-project/neo/issues/1672
        self.storagekey1 = storage.StorageKey(self.storage1_id, b'\x00\x01')
        self.storagekey2 = storage.StorageKey(self.storage1_id, b'\x00\x02')
        self.storagekey3 = storage.StorageKey(self.storage1_id, b'\x00\x03')
        self.storagekey4 = storage.StorageKey(self.storage1_id, b'\x00\x04')

        self.storageitem1 = storage.StorageItem(b'\x01\x01')
        self.storageitem2 = storage.StorageItem(b'\x02\x02')
        self.storageitem3 = storage.StorageItem(b'\x03\x03')
        self.storageitem4 = storage.StorageItem(b'\x04\x04')

        # prepare
        snapshot = self.db.get_snapshotview()
        snapshot.storages.put(self.storagekey1, self.storageitem1)

        raw = self.db.get_rawview()
        raw.storages.put(self.storagekey2, self.storageitem2)
        raw.storages.put(self.storagekey3, self.storageitem3)
        raw.storages.put(self.storagekey4, self.storageitem4)

        # test
        find_prefix = storage.create_find_prefix(self.storage1_id, b'\x00')
        iter = snapshot.storages.find(find_prefix)
        kv_pair = next(iter)
        self.assertEqual(self.storagekey1, kv_pair[0])

        kv_pair = snapshot.storages.get(self.storagekey3)
        kv_pair = next(iter)
        self.assertEqual(self.storagekey2, kv_pair[0])
        self.assertEqual(self.storageitem2, kv_pair[1])

        kv_pair = next(iter)
        self.assertEqual(self.storagekey3, kv_pair[0])
        self.assertEqual(self.storageitem3, kv_pair[1])

        kv_pair = next(iter)
        self.assertEqual(self.storagekey4, kv_pair[0])
        self.assertEqual(self.storageitem4, kv_pair[1])

    def test_issue_2572(self):
        snapshot = self.db.get_snapshotview()
        s1 = snapshot.clone()
        key2 = storage.StorageKey(0, b'\x01\x01')
        s1.storages.put(storage.StorageKey(0, b'\x00\x01'), storage.StorageItem(b'\x00'))
        s1.storages.put(key2, storage.StorageItem(b'\x05'))
        s1.storages.commit()
        item = s1.storages.get(key2, read_only=False)
        item.value = b'\x06'
        item2 = snapshot.storages.get(key2, read_only=True)
        self.assertEqual(b'\x05', item2.value)


class AbstractTransactionStorageTest(abc.ABC, unittest.TestCase):
    """
    A helper class to easily test backend specific code
    """

    @abc.abstractmethod
    def db_factory(self):
        """ Implement to return an instance of your DB """

    def setUp(self) -> None:
        self.db = self.db_factory()

        cosigner = payloads.Signer(account=types.UInt160.from_string("d7678dd97c000be3f33e9362e673101bac4ca654"),
                                   scope=payloads.WitnessScope.GLOBAL)

        witness = payloads.Witness(invocation_script=b'', verification_script=b'\x55')

        self.tx1 = payloads.Transaction(version=0,
                                        nonce=123,
                                        system_fee=456,
                                        network_fee=789,
                                        valid_until_block=1,
                                        attributes=[],
                                        signers=[cosigner],
                                        script=b'\x01\x02',
                                        witnesses=[witness])

        self.tx1_hash = self.tx1.hash()
        # by changing the script we change the TX hash
        self.tx2 = deepcopy(self.tx1)
        self.tx2.script = b'\x03\x04'
        self.tx2_hash = self.tx2.hash()

    def test_raw(self):
        raw_view = self.db.get_rawview()

        # we should not find anything in an empty db
        target_tx_hash = types.UInt256.zero()
        with self.assertRaises(KeyError) as context:
            raw_view.transactions.get(target_tx_hash)
        self.assertIsNone(raw_view.transactions.try_get(target_tx_hash))

        # fill the db
        raw_view.transactions.put(self.tx1)
        # and test it is immediately added
        tx_from_db = raw_view.transactions.try_get(self.tx1_hash)
        self.assertIsNotNone(tx_from_db)
        self.assertEqual(self.tx1, tx_from_db)

        # test getting all transactions
        raw_view.transactions.put(self.tx2)
        txs = list(raw_view.transactions.all())
        self.assertEqual(2, len(txs))
        self.assertIn(self.tx1, txs)
        self.assertIn(self.tx2, txs)

        # finally try removing the tx
        raw_view.transactions.delete(self.tx1_hash)
        self.assertIsNone(raw_view.transactions.try_get(self.tx1_hash))

    def test_snapshot_basic_add_delete_get(self):
        # test basic add, delete, get and separation
        raw_view = self.db.get_rawview()
        snapshot_view = self.db.get_snapshotview()

        # we should not find anything in an empty db
        target_tx_hash = types.UInt256.zero()
        with self.assertRaises(KeyError) as context:
            snapshot_view.transactions.get(target_tx_hash)
        self.assertIsNone(snapshot_view.transactions.try_get(target_tx_hash))

        # add item
        snapshot_view.transactions.put(self.tx1)
        # real backend should not be affected until a commit is called
        self.assertIsNone(raw_view.transactions.try_get(self.tx1_hash))

        # persist to backend
        snapshot_view.commit()
        tx_from_db = raw_view.transactions.try_get(self.tx1_hash)
        # and validate
        self.assertIsNotNone(tx_from_db)
        self.assertEqual(self.tx1, tx_from_db)

        # finally, try deleting
        # get a clean view with no cache
        snapshot_view = self.db.get_snapshotview()
        snapshot_view.transactions.delete(self.tx1_hash)

        # real backend should still have it, snapshot not
        self.assertIsNotNone(raw_view.transactions.try_get(self.tx1_hash))
        self.assertIsNone(snapshot_view.transactions.try_get(self.tx1_hash))

        # persist and validate real backend also doesn't have it anymore
        snapshot_view.commit()
        self.assertIsNone(raw_view.transactions.try_get(self.tx1_hash))

    def test_snapshot_add_duplicates(self):
        snapshot_view = self.db.get_snapshotview()

        # test double adding while already in cache
        snapshot_view.transactions.put(self.tx1)
        with self.assertRaises(ValueError):
            snapshot_view.transactions.put(self.tx1)

        # test double adding when not in cache, but in real backend
        snapshot_view.commit()
        # get a clean one with an empty cache
        snapshot_view = self.db.get_snapshotview()
        with self.assertRaises(ValueError):
            snapshot_view.transactions.put(self.tx1)

    def test_snapshot_add_while_cache_marked_deleted(self):
        # an item can exist in the real backend, and be marked in cache to be deleted
        # it should then be possible to delete it from cache without exceptions

        # fill real backend
        raw_view = self.db.get_rawview()
        raw_view.transactions.put(self.tx1)

        # ensure item is marked as deleted in cache
        snapshot_view = self.db.get_snapshotview()
        snapshot_view.transactions.delete(self.tx1_hash)

        # now test by adding add the item again
        success = False
        with suppress(ValueError):
            snapshot_view.transactions.put(self.tx1)
            success = True
        self.assertTrue(success)

    def test_snapshot_delete_various(self):
        snapshot_view = self.db.get_snapshotview()

        # delete non existing item with empty cache should throw no errors
        ok = False
        with suppress(Exception):
            snapshot_view.transactions.delete(self.tx1_hash)
            ok = True
        self.assertTrue(ok)

        # delete an item that was only added to the cache
        snapshot_view.transactions.put(self.tx1)
        snapshot_view.transactions.delete(self.tx1_hash)
        # we test if the cache is empty by looking at the count of items in the cache
        self.assertEqual(0, len(snapshot_view.transactions._dictionary))

        # finally we get an existing item, which puts it into CHANGED state
        # then DELETE it and verify it is marked as deleted

        # start by filling the real backend
        raw_view = self.db.get_rawview()
        raw_view.transactions.put(self.tx1)

        # get the item so it creates a cache entry
        tx_from_db = snapshot_view.transactions.get(self.tx1_hash)
        self.assertIsNotNone(tx_from_db)
        # now delete it
        snapshot_view.transactions.delete(self.tx1_hash)
        # and validate the item in cache is marked as DELETED
        self.assertEqual(1, len(snapshot_view.transactions._dictionary))
        trackable = snapshot_view.transactions._dictionary.get(self.tx1_hash, None)  # type: storage.Trackable
        self.assertIsNotNone(trackable)
        self.assertEqual(storage.TrackState.DELETED, trackable.state)

    def test_snapshot_get_various(self):
        snapshot_view = self.db.get_snapshotview()

        # get non existing item
        with self.assertRaises(KeyError):
            snapshot_view.transactions.get(self.tx1_hash)

        # test read only
        raw_view = self.db.get_rawview()
        raw_view.transactions.put(self.tx1)
        tx = snapshot_view.transactions.get(self.tx1_hash, read_only=True)
        tx.script = b'\x11\x22'

        tx_again = snapshot_view.transactions.get(self.tx1_hash, read_only=True)
        # We validate the hash of the original with the hash of the tx we retrieved.
        # The modification of the script attribute above changes the hash, if it persisted
        # the following test fails
        self.assertEqual(self.tx1_hash, tx_again.hash())

    def test_snapshot_clone_put(self):
        raw_view = self.db.get_rawview()
        snapshot_view = self.db.get_snapshotview()
        snapshot_view.transactions.put(self.tx1)

        clone_view = snapshot_view.clone()
        # validate it has the same tx
        self.assertEqual(self.tx1, clone_view.transactions.try_get(self.tx1_hash))

        # put some in the clone
        clone_view.transactions.put(self.tx2)

        # validate it is not visible anywhere but in the clone
        self.assertIsNone(raw_view.transactions.try_get(self.tx2_hash))
        self.assertIsNone(snapshot_view.transactions.try_get(self.tx2_hash))
        self.assertEqual(self.tx2, clone_view.transactions.try_get(self.tx2_hash))

        # commit changes of the clone back into the snapshot
        # should affect only the snapshot, not the real backend
        clone_view.commit()
        self.assertIsNone(raw_view.transactions.try_get(self.tx2_hash))
        self.assertEqual(self.tx2, snapshot_view.transactions.try_get(self.tx2_hash))

        # finally commit to real db
        snapshot_view.commit()
        self.assertEqual(self.tx2, raw_view.transactions.try_get(self.tx2_hash))

    def test_snapshot_clone_delete(self):
        raw_view = self.db.get_rawview()
        raw_view.transactions.put(self.tx1)
        snapshot_view = self.db.get_snapshotview()

        # perform a get to fill the cache
        snapshot_view.transactions.get(self.tx1_hash)

        clone_view = snapshot_view.clone()
        # now test deleting an item
        clone_view.transactions.delete(self.tx1_hash)
        # test it's gone in the clone, but nowhere else
        self.assertIsNone(clone_view.transactions.try_get(self.tx1_hash))
        self.assertIsNotNone(snapshot_view.transactions.try_get(self.tx1_hash))
        self.assertIsNotNone(raw_view.transactions.try_get(self.tx1_hash))

        # commit the clone into the snapshot
        clone_view.commit()
        # and validate it is also gone in the snapshot but not the real db
        self.assertIsNone(snapshot_view.transactions.try_get(self.tx1_hash))
        self.assertIsNotNone(raw_view.transactions.try_get(self.tx1_hash))

        # finally persist to real db
        snapshot_view.commit()
        self.assertIsNone(raw_view.transactions.try_get(self.tx1_hash))

    def test_snapshot_clone_update(self):
        # special note, we cannot change any of the official Transaction object attributes as that would change
        # the hash of the tx which is used as the identifier to find the item in the cache
        raw_view = self.db.get_rawview()
        raw_view.transactions.put(self.tx1)
        snapshot_view = self.db.get_snapshotview()

        # perform a get to fill the cache
        snapshot_view.transactions.get(self.tx1_hash)

        clone_view = snapshot_view.clone()
        tx_from_clone = clone_view.transactions.get(self.tx1_hash)  # type: payloads.Transaction
        # modify one of augmented attributes
        tx_from_clone.block_height = 1

        # validate the snapshot and real backend are not affected
        tx_from_snapshot = snapshot_view.transactions.get(self.tx1_hash)
        tx_from_real_db = raw_view.transactions.get(self.tx1_hash)
        self.assertNotEqual(1, tx_from_snapshot.block_height)
        self.assertNotEqual(1, tx_from_real_db.block_height)

        # commit clone
        clone_view.commit()
        # now snapshot should be updated, but real db not
        tx_from_snapshot = snapshot_view.transactions.get(self.tx1_hash)
        tx_from_real_db = raw_view.transactions.get(self.tx1_hash)
        self.assertEqual(1, tx_from_snapshot.block_height)
        self.assertNotEqual(1, tx_from_real_db.block_height)

        # finally persist to real db
        snapshot_view.commit()
        tx_from_real_db = raw_view.transactions.get(self.tx1_hash)
        self.assertEqual(1, tx_from_real_db.block_height)

    def test_all(self):
        raw_view = self.db.get_rawview()
        raw_view.transactions.put(self.tx1)
        raw_view.transactions.put(self.tx2)

        snapshot_view = self.db.get_snapshotview()

        # get() a tx to fill the cache so we can test sorting and readonly behaviour
        # tx2's hash comes before tx1 when sorting. So we cache that first as the all() internals
        # collect the results from the backend (=tx1) before results from the cache (=tx2).
        # Therefore if tx2 is found in the first position of the all() results, we can
        # conclude that the sort() happened correctly.
        snapshot_view.transactions.get(self.tx2_hash, read_only=True)

        txs = list(snapshot_view.transactions.all())
        self.assertEqual(2, len(txs))
        self.assertEqual(self.tx2, txs[0])
        self.assertEqual(self.tx1, txs[1])

        # ensure all() results are readonly
        txs[0].block_height = 999
        txs[1].block_height = 999

        tx1_from_snap = snapshot_view.transactions.get(self.tx1_hash, read_only=True)
        tx2_from_snap = snapshot_view.transactions.get(self.tx2_hash, read_only=True)
        self.assertNotEqual(999, tx1_from_snap.block_height)
        self.assertNotEqual(999, tx2_from_snap.block_height)

        # test clone all()
        tx3 = deepcopy(self.tx1)
        tx3.script = b'\x05\x06'
        clone_view = snapshot_view.clone()
        clone_view.transactions.put(tx3)

        txs = list(clone_view.transactions.all())
        self.assertEqual(3, len(txs))
        self.assertEqual(2, len(list(snapshot_view.transactions.all())))
        self.assertEqual(self.tx1, txs[2])
        self.assertEqual(self.tx2, txs[0])
        self.assertEqual(tx3, txs[1])



import unittest
from neo3 import storage, vm
from neo3.core import serialization, types


class StorageKeyTest(unittest.TestCase):
    def test_eq(self):
        contract = 1
        key_val = b'\0x01\x02\x03'
        sk = storage.StorageKey(contract, key_val)
        sk2 = storage.StorageKey(contract, key_val)
        self.assertFalse(sk == object())
        self.assertTrue(sk == sk2)

    def test_len(self):
        contract = 1
        key_val = b'\x01\x02\x03'
        sk = storage.StorageKey(contract, key_val)

        # contract id is serialized to int32
        expected_len = 4 + len(key_val)
        self.assertEqual(expected_len, len(sk))

    def test_serialization(self):
        contract = 1
        key_val = b'\x01\x02\x03'
        sk = storage.StorageKey(contract, key_val)

        # test serialize
        expected_value = b'\x01\x00\x00\x00' + key_val
        self.assertEqual(expected_value, sk.to_array())
        # test deserialize
        self.assertEqual(sk, storage.StorageKey.deserialize_from_bytes(expected_value))

    def test_addition(self):
        sk = storage.StorageKey(1, b'\x01')
        new_sk = sk + b'\x02'
        self.assertNotEqual(id(sk), id(new_sk))
        self.assertNotEqual(sk.key, new_sk.key)
        self.assertEqual(new_sk.key, b'\x01\x02')

        # test with serializable type
        new_sk2 = sk + types.UInt160.zero()
        self.assertEqual(new_sk2.key, b'\x01' + b'\x00' * 20)

        with self.assertRaises(TypeError) as context:
            sk + 1
        self.assertEqual("unsupported operand type(s) for +: 'StorageKey' and 'int'", str(context.exception))

    def test_various(self):
        contract = 1
        key_val = b'\x01\x02\x03'
        sk = storage.StorageKey(contract, key_val)
        # test __repr__ in absence of __str__
        self.assertIn("<StorageKey at ", str(sk))
        self.assertIn(r"[1] b'\x01\x02\x03'", str(sk))

        # test __hash__
        self.assertEqual(2161234437, hash(sk))


class StorageItemTest(unittest.TestCase):
    def test_eq(self):
        si = storage.StorageItem(b'\x01')
        si2 = storage.StorageItem(b'\x01')
        self.assertFalse(si == object())
        self.assertTrue(si == si2)

    def test_len(self):
        si = storage.StorageItem(b'\x01')
        self.assertEqual(2, len(si))

    def test_serialization(self):
        si_data = b'\x01\x02\x03'
        si = storage.StorageItem(si_data)
        self.assertEqual(si_data, si.to_array())
        self.assertEqual(si, storage.StorageItem.deserialize_from_bytes(si_data))

    def test_clone_from_replica(self):
        si_data = b'\x01\x02\x03'
        si = storage.StorageItem(si_data)
        clone = si.clone()
        self.assertEqual(si, clone)
        self.assertNotEqual(id(si), id(clone))

        si2 = storage.StorageItem(bytearray())
        si2.from_replica(si)
        self.assertEqual(si, si2)

    def test_getting_serializable(self):
        raw_value = b'\x01\x01'
        si = storage.StorageItem(raw_value)
        obj = si.get(TestSerializable)
        self.assertEqual(str(vm.BigInteger(1)), str(obj.value))
        self.assertEqual(raw_value, si.value)

        new_raw_value = b'\x01\x02'
        obj.value += 1
        obj2 = si.get(TestSerializable)
        self.assertEqual(id(obj), id(obj2))
        self.assertEqual(vm.BigInteger(2), obj2.value)
        self.assertEqual(new_raw_value, si.value)


class TestSerializable(serialization.ISerializable):
    def __init__(self):
        self.value = vm.BigInteger(1)

    def __len__(self):
        return 4

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        writer.write_var_bytes(self.value.to_array())

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        self.value = vm.BigInteger(reader.read_var_bytes())



import unittest
from neo3.core import utils, serialization
from neo3.core.serialization import BinaryReader, BinaryWriter
from enum import IntEnum


class DummyEnum(IntEnum):
    DEFAULT = 0


class DummySerializable(serialization.ISerializable):

    def serialize(self, writer: BinaryWriter) -> None:
        pass

    def deserialize(self, reader: BinaryReader) -> None:
        pass

    def __len__(self):
        return 2


class VarSizeTestCase(unittest.TestCase):
    def test_varsize_int(self):
        self.assertEqual(1, utils.get_var_size(0xFC))
        self.assertEqual(3, utils.get_var_size(0xFD))
        self.assertEqual(3, utils.get_var_size(0xFFFF))
        self.assertEqual(5, utils.get_var_size(0xFFFF + 1))

    def test_varsize_string(self):
        input = "abc"
        self.assertEqual(1+len(input), utils.get_var_size(input))

        input = "a" * 0xFC
        self.assertEqual(1 + len(input), utils.get_var_size(input))

        # boundary check
        input = "a" * 0xFD
        self.assertEqual(3 + len(input), utils.get_var_size(input))

        input = "a" * 0xFFFF
        self.assertEqual(3 + len(input), utils.get_var_size(input))

        input = "a" * (0xFFFF + 1)
        self.assertEqual(5 + len(input), utils.get_var_size(input))

    def test_iterables(self):
        iterable = []
        self.assertEqual(1, utils.get_var_size(iterable))

        iterable = b'\x01\x02'
        self.assertEqual(1 + len(iterable), utils.get_var_size(iterable))

        iterable = [DummySerializable(), DummySerializable()]
        fixed_dummy_size = 2
        self.assertEqual(1 + (2 * fixed_dummy_size), utils.get_var_size(iterable))

        # so far NEO only has byte enums, so the length is fixed to 1
        iterable = [DummyEnum.DEFAULT, DummyEnum.DEFAULT]
        self.assertEqual(1 + 2, utils.get_var_size(iterable))

        # test unsupported type in iterable
        iterable = [object()]
        with self.assertRaises(TypeError) as context:
            utils.get_var_size(iterable)
        self.assertIn("Cannot accurately determine size of objects that do not", str(context.exception))

    def test_not_supported_objects(self):
        with self.assertRaises(ValueError) as context:
            utils.get_var_size(object())
        self.assertIn("NOT SUPPORTED", str(context.exception))



import unittest
from neo3 import storage


class ComparatorTestCase(unittest.TestCase):
    def test_forward(self):
        comperator = storage.NEOByteCompare("forward")
        self.assertEqual(-1, comperator.compare(b'\x00', b'\x01'))
        self.assertEqual(0, comperator.compare(b'\x01', b'\x01'))
        self.assertEqual(1, comperator.compare(b'\x02', b'\x01'))
        self.assertEqual(-1, comperator.compare(b'\x01', b'\x02'))
        self.assertEqual(1, comperator.compare(b'\x01\x01', b'\x01'))
        self.assertEqual(1, comperator.compare(b'\xff', b'\x01'))
        self.assertEqual(-1, comperator.compare(b'\xf9', b'\xfb'))
        self.assertEqual(1, comperator.compare(b'\xf9\xff\xff\xff\x09', b'\x01\x00\x00\x00\x01'))
        self.assertEqual(1, comperator.compare(b'\x01\x00\x00\x00\x01\x02', b'\x01\x00\x00\x00\x01'))




from __future__ import annotations
import hashlib
import abc
import struct
import base58  # type: ignore
import base64
from enum import Enum
from typing import List, Optional, Type, TypeVar
from neo3.core import Size as s, serialization, utils, types, IInteroperable, IJson
from neo3.network import payloads
from neo3.vm import VMState
from neo3 import settings, vm, storage, contracts


class TransactionAttributeType(Enum):
    HIGH_PRIORITY = 0x1
    ORACLE_RESPONSE = 0x11

    def to_csharp_name(self) -> str:
        """
        Internal helper to match C# convention
        """
        if self == self.HIGH_PRIORITY:
            return "HighPriority"
        else:
            return "OracleResponse"

    @classmethod
    def from_csharp_name(cls, name: str):
        """
        Internal helper to parse from C# convention
        """
        if name == "HighPriority":
            return cls.HIGH_PRIORITY
        else:
            return cls.ORACLE_RESPONSE


TransactionAttribute_T = TypeVar('TransactionAttribute_T', bound='TransactionAttribute')


class TransactionAttribute(serialization.ISerializable, IJson):
    """
    Attributes that can be attached to a Transaction.
    """

    def __init__(self):
        self.type_: TransactionAttributeType = None
        self.allow_multiple = False

    def __len__(self):
        return s.uint8

    def __eq__(self, other):
        if other is None:
            return False
        if type(self) != type(other):
            return False
        return True

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        """
        Serialize the object into a binary stream.

        Args:
            writer: instance.
        """
        writer.write_uint8(self.type_.value)
        self._serialize_without_type(writer)

    @abc.abstractmethod
    def _serialize_without_type(self, writer: serialization.BinaryWriter) -> None:
        """ Serialize the remaining attributes"""

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        """
        Deserialize the object from a binary stream.

        Args:
            reader: instance.
        """
        if reader.read_uint8() != self.type_.value:
            raise ValueError("Deserialization error - transaction attribute type mismatch")
        self._deserialize_without_type(reader)

    @staticmethod
    def deserialize_from(reader: serialization.BinaryReader) -> TransactionAttribute:
        """
        Deserialize from a binary stream into a new TransactionAttribute
        """
        attribute_type = reader.read_uint8()
        for sub in TransactionAttribute.__subclasses__():
            child = sub._serializable_init()  # type: ignore
            if child.type_.value == attribute_type:
                child._deserialize_without_type(reader)
                return child
        else:
            raise ValueError("Deserialization error - unknown transaction attribute type")

    @abc.abstractmethod
    def _deserialize_without_type(self, reader: serialization.BinaryReader) -> None:
        """ Deserialize the remaining attributes """

    def verify(self, snapshot: storage.Snapshot, tx: Transaction) -> bool:
        return True

    def to_json(self) -> dict:
        """ Convert object into json """
        return {"type": self.type_.to_csharp_name()}

    @classmethod
    def from_json(cls, json: dict):
        """ Create object from JSON """
        c = cls()
        c.type_ = TransactionAttributeType(json["type"])


class HighPriorityAttribute(TransactionAttribute):
    def __init__(self):
        super(HighPriorityAttribute, self).__init__()
        self.type_ = TransactionAttributeType.HIGH_PRIORITY

    def verify(self, snapshot: storage.Snapshot, tx: Transaction) -> bool:
        """
        Verifies the attribute with the transaction

        Returns:
            True if verification passes. False otherwise.
        """
        committee = contracts.NeoToken().get_committee_address(snapshot)
        for signer in tx.signers:
            if signer.account == committee:
                return True
        return False

    def _serialize_without_type(self, writer: serialization.BinaryWriter) -> None:
        pass

    def _deserialize_without_type(self, reader: serialization.BinaryReader) -> None:
        pass


class Transaction(payloads.IInventory, IInteroperable, IJson):
    """
    Data to be executed by the NEO virtual machine.
    """
    #: the maximum number of bytes a single transaction may consists of
    MAX_TRANSACTION_SIZE = 102400
    #: the maximum time a transaction will be valid from height of creation plus this value. Default is 24h
    MAX_VALID_UNTIL_BLOCK_INCREMENT = 5760
    #: the maximum number of transaction attributes for a single transaction
    MAX_TRANSACTION_ATTRIBUTES = 16

    HEADER_SIZE = (1  # Version
                   + 4  # NONCE
                   + 8  # SYSTEM_FEE
                   + 8  # NETWORK_FEE
                   + 4)  # VALID_UNTIL_BLOCK

    def __init__(self,
                 version: int,
                 nonce: int,
                 system_fee: int,
                 network_fee: int,
                 valid_until_block: int,
                 attributes: List[TransactionAttribute] = None,
                 signers: List[payloads.Signer] = None,
                 script: bytes = None,
                 witnesses: List[payloads.Witness] = None,
                 protocol_magic: int = None):
        #: Transaction data structure version - for internal use
        self.version = version
        #: Random number
        self.nonce = nonce
        #: The cost of executing the `script`.
        self.system_fee = system_fee
        #: The cost of validation and inclusion in a block by the consensus node.
        self.network_fee = network_fee
        #: The maximum chain height this transaction is valid.
        self.valid_until_block = valid_until_block
        """
        Optional attributes

        See also:
            :class:`~neo3.network.payloads.transaction.HighPriorityAttribute`.
            :class:`~neo3.network.payloads.oracle.OracleResponse`.
        """
        self.attributes = attributes if attributes else []
        #: A list of authorities used by the :func:`ChecKWitness` smart contract system call.
        self.signers = signers if signers else []
        #: The array of instructions to be executed on the chain by the virtual machine.
        self.script = script if script else b''
        #: A list of signing authorities used to validate the transaction.
        self.witnesses = witnesses if witnesses else []

        # unofficial attributes
        #: The virtual machine result of executing the `script`.
        self.vm_state = VMState.NONE
        #: The block height in which the transaction is included.
        self.block_height = 0
        #: The network protocol magic number to use in the Transaction hash function. Defaults to 0x4F454E
        #: Warning: changing this will change the TX hash which can result in dangling transactions in the database as
        #: deletion and duplication checking will fail.
        if protocol_magic:
            self.protocol_magic = protocol_magic
        elif settings.network.magic is not None:
            self.protocol_magic = settings.network.magic
        else:
            self.protocol_magic = 0x4F454E

    def __len__(self):
        return (s.uint8 + s.uint32 + s.uint64 + s.uint64 + s.uint32
                + utils.get_var_size(self.attributes)
                + utils.get_var_size(self.signers)
                + utils.get_var_size(self.script)
                + utils.get_var_size(self.witnesses))

    def __eq__(self, other):
        if other is None:
            return False
        if type(self) != type(other):
            return False
        if self.hash() != other.hash():
            return False
        return True

    def __hash__(self):
        # the TX hash() is a UInt256, we need to return an int
        # so we call hash() on the UInt256
        return hash(self.hash())

    def __deepcopy__(self, memodict={}):
        # not the best, but faster than letting deepcopy() do introspection
        with serialization.BinaryWriter() as bw:
            self.serialize_special(bw)
            with serialization.BinaryReader(bw.to_array()) as br:
                tx = Transaction._serializable_init()
                tx.deserialize_special(br)
                return tx

    def hash(self) -> types.UInt256:
        """
        Get a unique block identifier based on the unsigned data portion of the object.
        """
        with serialization.BinaryWriter() as bw:
            self.serialize_unsigned(bw)
            data_to_hash = bytearray(bw._stream.getvalue())
            data = hashlib.sha256(data_to_hash).digest()
            return types.UInt256(data=data)

    @property
    def sender(self) -> types.UInt160:
        """
        The hash of the account who has send the transaction to the network
        """
        return self.signers[0].account if len(self.signers) > 0 else types.UInt160.zero()

    @property
    def inventory_type(self) -> payloads.InventoryType:
        """
        Inventory type identifier.
        """
        return payloads.InventoryType.TX

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        """
        Serialize the object into a binary stream.

        Args:
            writer: instance.
        """
        self.serialize_unsigned(writer)
        writer.write_serializable_list(self.witnesses)

    def serialize_unsigned(self, writer: serialization.BinaryWriter) -> None:
        """
        Serialize the unsigned part of the object into a binary stream.

        Args:
            writer: instance.
        """
        writer.write_uint8(self.version)
        writer.write_uint32(self.nonce)
        writer.write_int64(self.system_fee)
        writer.write_int64(self.network_fee)
        writer.write_uint32(self.valid_until_block)
        writer.write_serializable_list(self.signers)
        writer.write_serializable_list(self.attributes)
        writer.write_var_bytes(self.script)

    def serialize_special(self, writer: serialization.BinaryWriter) -> None:
        """ Internal use only - serialize the TX includes its unofficial fields """
        self.serialize(writer)
        writer.write_uint8(int(self.vm_state))
        writer.write_uint32(self.block_height)

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        """
        Deserialize the object from a binary stream.

        Args:
            reader: instance.
        """
        self.deserialize_unsigned(reader)
        self.witnesses = reader.read_serializable_list(payloads.Witness, max=len(self.signers))
        if len(self.witnesses) != len(self.signers):
            raise ValueError("Deserialization error - witness length does not match signers length")

    def deserialize_unsigned(self, reader: serialization.BinaryReader) -> None:
        """
        Deserialize the unsigned data part of the object from a binary stream.

        Args:
            reader: instance.

        Raises:
            ValueError: The version is not zero.
            ValueError: If the system of network fee is negative.
            ValueError: If there is no script
        """
        (self.version,
         self.nonce,
         self.system_fee,
         self.network_fee,
         self.valid_until_block) = struct.unpack("<BIqqI", reader._stream.read(25))
        if self.version > 0:
            raise ValueError("Deserialization error - invalid version")
        if self.system_fee < 0:
            raise ValueError("Deserialization error - negative system fee")
        if self.network_fee < 0:
            raise ValueError("Deserialization error - negative network fee")

        self.signers = Transaction._deserialize_signers(reader, self.MAX_TRANSACTION_ATTRIBUTES)
        self.attributes = Transaction._deserialize_attributes(reader,
                                                              self.MAX_TRANSACTION_ATTRIBUTES - len(self.signers))
        self.script = reader.read_var_bytes(max=65535)
        if len(self.script) == 0:
            raise ValueError("Deserialization error - invalid script length 0")

    def deserialize_special(self, reader: serialization.BinaryReader) -> None:
        """ Internal use only - deserialize the data from the stream into a TX that includes the unofficial fields """
        self.deserialize(reader)
        self.vm_state = VMState(reader.read_uint8())
        self.block_height = reader.read_uint32()

    def fee_per_byte(self) -> int:
        """
        Calculates the network fee per byte.

        Fee per byte = the TX's networkfee / TX's size

        Warning:
            Should only be called once the transaction is completely build and will no longer be modified.
        """
        return self.network_fee // len(self)

    def from_replica(self, replica) -> None:
        """
        Shallow copy attributes from a reference object.
        """
        self.version = replica.version
        self.nonce = replica.nonce
        self.system_fee = replica.system_fee
        self.network_fee = replica.network_fee
        self.valid_until_block = replica.valid_until_block
        self.attributes = replica.attributes
        self.signers = replica.signers
        self.script = replica.script
        self.witnesses = replica.witnesses
        self.block_height = replica.block_height
        self.vm_state = replica.vm_state

    def to_stack_item(self, reference_counter: vm.ReferenceCounter) -> vm.StackItem:
        """
        Convert self to a VM stack item.

        Args:
            reference_counter: ExecutionEngine reference counter
        """
        array = vm.ArrayStackItem(reference_counter)
        tx_hash = vm.ByteStringStackItem(self.hash().to_array())
        version = vm.IntegerStackItem(self.version)
        nonce = vm.IntegerStackItem(self.nonce)
        sender = vm.ByteStringStackItem(self.sender.to_array())
        system_fee = vm.IntegerStackItem(self.system_fee)
        network_fee = vm.IntegerStackItem(self.network_fee)
        valid_until = vm.IntegerStackItem(self.valid_until_block)
        script = vm.ByteStringStackItem(self.script)
        array.append([tx_hash, version, nonce, sender, system_fee, network_fee, valid_until, script])
        return array

    def to_json(self) -> dict:
        """ Convert object into json """
        version = b'\x35'
        # replace with to_address once the feature-wallet branch is merged
        x = version + self.sender.to_array()
        sender = base58.b58encode_check(x).decode()
        return {
            "hash": "0x" + str(self.hash()),
            "size": len(self),
            "version": self.version,
            "nonce": self.nonce,
            "sender": sender,
            "sysfee": str(self.system_fee),
            "netfee": str(self.network_fee),
            "validuntilblock": self.valid_until_block,
            "signers": list(map(lambda s: s.to_json(), self.signers)),
            "attributes": list(map(lambda a: a.to_json(), self.attributes)),
            "script": base64.b64encode(self.script).decode(),
            "witnesses": list(map(lambda w: w.to_json(), self.witnesses))
        }

    @classmethod
    def from_json(cls, json: dict, procotol_magic=None):
        """ Create object from JSON """
        version = json['version']
        nonce = json['nonce']
        system_fee = int(json['sysfee'])
        network_fee = int(json['netfee'])
        valid_until_block = json['validuntilblock']
        attributes: List[TransactionAttribute] = []
        # ugh :-(
        for attribute in json['attributes']:
            try:
                type_ = payloads.TransactionAttributeType.from_csharp_name(attribute['type'])
                if type_ == payloads.TransactionAttributeType.HIGH_PRIORITY:
                    attributes.append(payloads.HighPriorityAttribute())
                elif type_ == payloads.TransactionAttributeType.ORACLE_RESPONSE:
                    attributes.append(payloads.OracleResponse.from_json(attribute))
            except ValueError:
                raise ValueError("Invalid transaction attribute")
        signers = list(map(lambda s: payloads.Signer.from_json(s), json['signers']))
        script = base64.b64decode(json['script'].encode())
        witnesses = list(map(lambda w: payloads.Witness.from_json(w), json['witnesses']))
        return cls(version,
                   nonce,
                   system_fee,
                   network_fee,
                   valid_until_block,
                   attributes,
                   signers,
                   script,
                   witnesses,
                   procotol_magic)

    def get_script_hashes_for_verifying(self, _: storage.Snapshot) -> List[types.UInt160]:
        """
        Helper method to get the data used in verifying the object.
        """
        return list(map(lambda signer: signer.account, self.signers))

    def try_get_attribute(self, needle: Type[TransactionAttribute_T]) -> \
            Optional[TransactionAttribute_T]:
        """
        Helper method for finding an attribute of a specific type.

        Args:
            needle: the type to search for

        Returns:
            The attribute matching the type if found. None otherwise.
        """
        for attr in self.attributes:
            if isinstance(attr, needle):
                return attr
        else:
            return None

    # TODO: implement Verify methods once we have Snapshot support

    @staticmethod
    def _deserialize_signers(reader: serialization.BinaryReader, max_count: int) -> List[payloads.Signer]:
        count = reader.read_var_int(max_count)
        if count == 0:
            raise ValueError("Deserialization error - signers can't be empty")

        values: List[payloads.Signer] = []
        for i in range(0, count):
            signer = reader.read_serializable(payloads.Signer)
            if signer in values:
                raise ValueError("Deserialization error - duplicate signer")
            values.append(signer)

        return values

    @staticmethod
    def _deserialize_attributes(reader: serialization.BinaryReader, max_count: int) -> List[TransactionAttribute]:
        count = reader.read_var_int(max_count)
        values: List[TransactionAttribute] = []

        for _ in range(0, count):
            attribute = TransactionAttribute.deserialize_from(reader)
            if not attribute.allow_multiple and attribute in values:
                raise ValueError("Deserialization error - duplicate transaction attribute")
            values.append(attribute)
        return values

    @classmethod
    def _serializable_init(cls):
        return cls(0, 0, 0, 0, 99999)



from enum import IntFlag


class TriggerType(IntFlag):
    ON_PERSIST = 0x01
    POST_PERSIST = 0x02
    SYSTEM = ON_PERSIST | POST_PERSIST
    VERIFICATION = 0x20
    APPLICATION = 0x40
    ALL = SYSTEM | VERIFICATION | APPLICATION



from unittest import TestCase
from neo3.core.types.uint import _UIntBase
from neo3.core.types import UInt160, UInt256
from neo3.core import serialization


class UIntBase(_UIntBase):
    def serialize(self) -> bytearray:
        pass

    @classmethod
    def deserialize(cls, data: bytes):
        pass


class UIntBaseTest(TestCase):
    def test_create_with_empty_data(self):
        x = UIntBase(num_bytes=2)
        self.assertEqual(len(x._data), 2)
        self.assertEqual(x._data, b'\x00\x00')

    def test_valid_rawbytes_data(self):
        x = UIntBase(num_bytes=2, data=b'\xaa\xbb')
        self.assertEqual(len(x._data), 2)
        self.assertNotEqual(len(x._data), 4)

    def test_raw_data_that_can_be_decoded(self):
        """
        some raw data can be decoded e.g. bytearray.fromhex('1122') but shouldn't be
        """
        tricky_raw_data = bytes.fromhex('1122')
        x = UIntBase(num_bytes=2, data=tricky_raw_data)
        self.assertEqual(x._data, tricky_raw_data)

    def test_data_length_mistmatch(self):
        with self.assertRaises(ValueError) as context:
            x = UIntBase(num_bytes=2, data=b'a')  # 2 != 1
        self.assertTrue("Invalid UInt: data length" in str(context.exception))

    def test_size(self):
        x = UIntBase(num_bytes=2, data=b'\xaa\xbb')
        self.assertEqual(len(x), 2)

    def test_hash_code(self):
        x = UIntBase(num_bytes=4, data=bytearray.fromhex('DEADBEEF'))
        self.assertEqual(hash(x), 4022250974)
        x = UIntBase(num_bytes=2, data=bytearray.fromhex('1122'))
        self.assertEqual(hash(x), 8721)

    def test_to_string(self):
        x = UIntBase(num_bytes=2, data=bytearray.fromhex('1122'))
        self.assertEqual('2211', str(x))
        self.assertNotEqual('1122', str(x))

    def test_equal(self):
        x = UIntBase(num_bytes=2, data=bytearray.fromhex('1122'))
        y = UIntBase(num_bytes=2, data=bytearray.fromhex('1122'))
        z = UIntBase(num_bytes=2, data=bytearray.fromhex('2211'))

        self.assertFalse(x == None)
        self.assertFalse(x == int(1122))
        self.assertTrue(x == x)
        self.assertTrue(x == y)
        self.assertTrue(x != z)

    def test_hash(self):
        x = UIntBase(num_bytes=2, data=bytearray.fromhex('1122'))
        y = UIntBase(num_bytes=2, data=bytearray.fromhex('1122'))
        z = UIntBase(num_bytes=2, data=bytearray.fromhex('2211'))
        self.assertEqual(hash(x), hash(y))
        self.assertNotEqual(hash(x), hash(z))

    def test_compare_to(self):
        x = UIntBase(num_bytes=2, data=bytearray.fromhex('1122'))
        y = UIntBase(num_bytes=3, data=bytearray.fromhex('112233'))
        z = UIntBase(num_bytes=2, data=bytearray.fromhex('1133'))
        xx = UIntBase(num_bytes=2, data=bytearray.fromhex('1122'))

        # test invalid type
        with self.assertRaises(TypeError) as context:
            x._compare_to(None)

        expected = "Cannot compare UIntBase to type NoneType"
        self.assertEqual(expected, str(context.exception))

        # test invalid length
        with self.assertRaises(ValueError) as context:
            x._compare_to(y)

        expected = "Cannot compare UIntBase with length 2 to UIntBase with length 3"
        self.assertEqual(expected, str(context.exception))

        # test data difference ('22' < '33')
        self.assertEqual(-1, x._compare_to(z))
        # test data difference ('33' > '22')
        self.assertEqual(1, z._compare_to(x))
        # test data equal
        self.assertEqual(0, x._compare_to(xx))

    def test_rich_comparison_methods(self):
        x = UIntBase(num_bytes=2, data=bytearray.fromhex('1122'))
        z = UIntBase(num_bytes=2, data=bytearray.fromhex('1133'))
        xx = UIntBase(num_bytes=2, data=bytearray.fromhex('1122'))

        self.assertTrue(x < z)
        self.assertTrue(z > x)
        self.assertTrue(x <= xx)
        self.assertTrue(x >= xx)


class UInt160_and_256Test(TestCase):
    def test_zero(self):
        uint160 = UInt160.zero()
        self.assertEqual(20, len(uint160.to_array()))

        uint256 = UInt256.zero()
        self.assertEqual(32, len(uint256.to_array()))

    def test_from_string_wrong_length(self):
        with self.assertRaises(ValueError) as ctx:
            UInt160.from_string("1122")
        self.assertEqual("Invalid UInt160 Format: 4 chars != 40 chars", str(ctx.exception))

        with self.assertRaises(ValueError) as ctx:
            UInt256.from_string("1122")
        self.assertEqual("Invalid UInt256 Format: 4 chars != 64 chars", str(ctx.exception))

    def test_from_string_various(self):
        uint160 = UInt160.from_string("11" * 20)
        expected_data_uint160 = bytearray([0x11] * 20)
        self.assertEqual(expected_data_uint160, uint160.to_array())

        uint256 = UInt256.from_string("11" * 32)
        expected_data_uint256 = bytearray([0x11] * 32)
        self.assertEqual(expected_data_uint256, uint256.to_array())


        uint160_from_bytes = UInt160.deserialize_from_bytes(expected_data_uint160)
        self.assertEqual(expected_data_uint160, uint160_from_bytes.to_array())

        uint256_from_bytes = UInt256.deserialize_from_bytes(expected_data_uint256)
        self.assertEqual(expected_data_uint256, uint256_from_bytes.to_array())

        # test deserialize with too much data
        data_uint160 = bytearray(21 * [0x11])
        uint160_from_bytes = UInt160.deserialize_from_bytes(data_uint160)
        self.assertEqual(data_uint160[:20], uint160_from_bytes.to_array())

        data_uint256 = bytearray(33 * [0x11])
        uint256_from_bytes = UInt256.deserialize_from_bytes(data_uint256)
        self.assertEqual(expected_data_uint256[:32], uint256_from_bytes.to_array())

        # test deserialize with too little data
        data_uint160 = bytearray(19 * [0x11])
        data_uint256 = bytearray(31 * [0x11])
        with self.assertRaises(ValueError) as ctx:
            UInt160.deserialize_from_bytes(data_uint160)
        self.assertEqual("Insufficient data 19 bytes is less than the required 20", str(ctx.exception))

        with self.assertRaises(ValueError) as ctx:
            UInt256.deserialize_from_bytes(data_uint256)
        self.assertEqual("Insufficient data 31 bytes is less than the required 32", str(ctx.exception))

    def test_deserialize_from_stream(self):
        data_uint160 = bytearray(20 * [0x11])
        data_uint256 = bytearray(32 * [0x11])

        with serialization.BinaryReader(data_uint160) as br:
            # we explicitly call deserialize, instead of br.read_uint160() for coverage
            uint160 = UInt160.zero()
            uint160.deserialize(br)
            self.assertEqual(data_uint160, uint160._data)

        with serialization.BinaryReader(data_uint256) as br:
            uint256 = UInt256.zero()
            uint256.deserialize(br)
            self.assertEqual(data_uint256, uint256._data)

    def test_serialize_to_stream(self):
        data_uint160 = bytearray(20 * [0x11])
        data_uint256 = bytearray(32 * [0x11])
        uint160 = UInt160(data_uint160)
        uint256 = UInt256(data_uint256)

        with serialization.BinaryWriter() as bw:
            bw.write_serializable(uint160)
            self.assertEqual(data_uint160, bw._stream.getvalue())

        with serialization.BinaryWriter() as bw:
            bw.write_serializable(uint256)
            self.assertEqual(data_uint256, bw._stream.getvalue())



from __future__ import annotations
from typing import Type, Optional
from neo3.core import serialization, utils, IClonable, Size as s


class StorageItem(serialization.ISerializable, IClonable):
    def __init__(self, value: bytes):
        self._value = value
        self._cache: Optional[serialization.ISerializable] = None

    def __len__(self):
        return utils.get_var_size(self.value)

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        return self.value == other.value

    @property
    def value(self) -> bytes:
        if self._cache:
            return self._cache.to_array()
        return self._value

    @value.setter
    def value(self, new_value: bytes) -> None:
        self._value = new_value
        self._cache = None

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        """
        Serialize the object into a binary stream.

        Args:
            writer: instance.
        """
        writer.write_bytes(self.value)

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        """
        Deserialize the object from a binary stream.

        Args:
            reader: instance.
        """
        remaining_stream_size = len(reader) - reader._stream.tell()
        self.value = reader.read_bytes(remaining_stream_size)

    def clone(self) -> StorageItem:
        """ Deep clone """
        return StorageItem(self.value)

    def from_replica(self, replica: StorageItem) -> None:
        """ Copy instance variables from other instance """
        self.value = replica.value

    def get(self, type_: Type[serialization.ISerializable]):
        """ Transform the data into `type` and cache the value """
        if self._cache and type(self._cache) == type_:
            return self._cache

        t = type_._serializable_init()
        self._cache = t.deserialize_from_bytes(self.value)
        return self._cache

    @classmethod
    def _serializable_init(cls):
        return cls(b'')



import unittest
import binascii
from neo3 import contracts
from neo3.core import types
from copy import deepcopy


class NEFTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        """
        var nef = new NefFile
        {
            Compiler = "test-compiler 0.1",
            Source = "source_link",
            Script = new byte[] {(byte) OpCode.RET},
            Tokens = new MethodToken[]
            {
                new MethodToken()
                {
                    Hash = UInt160.Zero,
                    Method = "test_method",
                    ParametersCount = 0,
                    HasReturnValue = true,
                    CallFlags = CallFlags.None
                }
            }
        };
        nef.CheckSum = NefFile.ComputeChecksum(nef);
        Console.WriteLine(nef.ToArray().ToHexString());
        Console.WriteLine(nef.Size);
        """
        cls.expected = binascii.unhexlify(b'4e454633746573742d636f6d70696c657220302e3100000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000b736f757263655f6c696e6b000100000000000000000000000000000000000000000b746573745f6d6574686f64000001000000014072adaf87')
        cls.expected_length = 126
        compiler = "test-compiler 0.1"
        source = "source_link"
        ret = b'\x40'  # vm.OpCode.RET
        tokens = [contracts.MethodToken(types.UInt160.zero(), "test_method", 0, True, contracts.CallFlags.NONE)]
        cls.nef = contracts.NEF(compiler_name=compiler, script=ret, tokens=tokens, source=source)

    def test_serialization(self):
        self.assertEqual(self.expected, self.nef.to_array())

    def test_deserialization(self):
        nef = contracts.NEF.deserialize_from_bytes(self.expected)
        self.assertEqual(self.nef.magic, nef.magic)
        self.assertEqual(self.nef.source, nef.source)
        self.assertEqual(self.nef.compiler, nef.compiler)
        self.assertEqual(self.nef.script, nef.script)
        self.assertEqual(self.nef.checksum, nef.checksum)

    def test_deserialization_error(self):
        nef = deepcopy(self.nef)
        nef.magic = 0xDEADBEEF
        with self.assertRaises(ValueError) as context:
            contracts.NEF.deserialize_from_bytes(nef.to_array())
        self.assertEqual("Deserialization error - Incorrect magic", str(context.exception))

        nef = deepcopy(self.nef)
        nef.script = b''
        with self.assertRaises(ValueError) as context:
            contracts.NEF.deserialize_from_bytes(nef.to_array())
        self.assertEqual("Deserialization error - Script can't be empty", str(context.exception))

        # test with wrong checksum
        nef = deepcopy(self.nef)
        nef._checksum = 0xDEADBEEF
        with self.assertRaises(ValueError) as context:
            contracts.NEF.deserialize_from_bytes(nef.to_array())
        self.assertEqual("Deserialization error - Invalid checksum", str(context.exception))

    def test_len(self):
        self.assertEqual(self.expected_length, len(self.nef))

    def test_eq(self):
        compiler = "test-compiler 0.1"
        ret = b'\x40'  # vm.OpCode.RET
        nef = contracts.NEF(compiler_name=compiler, script=ret)
        nef2 = contracts.NEF(compiler_name=compiler, script=ret)
        self.assertFalse(nef == object())
        self.assertTrue(nef == nef2)
